@@ -21,6 +21,23 @@
 
 #include "includes.h"
 #include "realcalls.h"
+#include "smbw_xstat.h"
+
+#ifndef ENOSYS
+#define ENOSYS	(-1)
+#endif
+
+#if !defined(NO_ACL_WRAPPER) || !defined(NO_FACL_WRAPPER)
+#ifdef HPUX
+#define aclent_t	void
+#ifndef GETACL
+#define GETACL		0x01	/* as ACL_GET */
+#endif
+#ifndef GETACLCNT
+#define GETACLCNT	0x03	/* as ACL_CNT */
+#endif
+#endif
+#endif
 
 pstring smbw_cwd;
 
@@ -29,7 +46,7 @@ static struct smbw_server *smbw_srvs;
 
 struct bitmap *smbw_file_bmap;
 extern pstring global_myname;
-extern int DEBUGLEVEL;
+extern BOOL AllowDebugChange;
 
 fstring smbw_prefix = SMBW_PREFIX;
 
@@ -45,21 +62,25 @@ initialise structures
 void smbw_init(void)
 {
 	extern BOOL in_client;
-	static int initialised;
+	static pid_t initialised = 0;
 	static pstring servicesf = CONFIGFILE;
 	extern FILE *dbf;
+	pid_t pid;
 	char *p;
 	int eno;
 	pstring line;
 
-	if (initialised) return;
-	initialised = 1;
+	/* static variables are not always reset after fork() */
+	pid = getpid();
+	if (initialised == pid) return;
+	initialised = pid;
 
 	eno = errno;
 
 	smbw_busy++;
 
 	DEBUGLEVEL = 0;
+	AllowDebugChange = False;
 	setup_logging("smbsh",True);
 
 	dbf = stderr;
@@ -79,7 +100,12 @@ void smbw_init(void)
 
 	load_interfaces();
 
+	if ((p=smbw_getshared("SERVICESF"))) {
+		pstrcpy(servicesf, p);
+	}
+
 	lp_load(servicesf,True,False,False);
+	codepage_initialise(lp_client_code_page());
 
 	get_myname(global_myname);
 
@@ -97,13 +123,19 @@ void smbw_init(void)
 		DEBUG(2,("SMBW_PREFIX is %s\n", smbw_prefix));
 	}
 
+	if ((p=smbw_getshared("TERM_CODE"))) {
+		interpret_coding_system(p);
+	}
+
 	slprintf(line,sizeof(line)-1,"PWD_%d", (int)getpid());
 	
 	p = smbw_getshared(line);
 	if (!p) {
 		sys_getwd(smbw_cwd);
 	}
-	pstrcpy(smbw_cwd, p);
+	else {
+		pstrcpy(smbw_cwd, p);
+	}
 	DEBUG(4,("Initial cwd is %s\n", smbw_cwd));
 
 	smbw_busy--;
@@ -246,89 +278,111 @@ void clean_fname(char *name)
 }
 
 
+
+/***************************************************** 
+find a workgroup (any workgroup!) that has a master 
+browser on the local network
+*******************************************************/
+static char *smbw_find_workgroup(void)
+{
+	fstring server;
+	char *p;
+	struct in_addr *ip_list = NULL;
+	int count = 0;
+	int i;
+
+	/* first off see if an existing workgroup name exists */
+	p = smbw_getshared("WORKGROUP");
+	if (!p) p = lp_workgroup();
+	
+	slprintf(server, sizeof(server), "%s#1D", p);
+	if (smbw_server(server, "IPC$")) return p;
+
+	/* go looking for workgroups */
+	if (!name_resolve_bcast(MSBROWSE, 1, &ip_list, &count)) {
+		DEBUG(1,("No workgroups found!"));
+		return p;
+	}
+
+	for (i=0;i<count;i++) {
+		static fstring name;
+		if (name_status_find("*", 0, 0x1d, ip_list[i], name)) {
+			slprintf(server, sizeof(server), "%s#1D", name);
+			if (smbw_server(server, "IPC$")) {
+				smbw_setshared("WORKGROUP", name);
+				SAFE_FREE(ip_list);
+				return name;
+			}
+		}
+	}
+
+	SAFE_FREE(ip_list);
+
+	return p;
+}
+
 /***************************************************** 
 parse a smb path into its components. 
+server is one of
+  1) the name of the SMB server
+  2) WORKGROUP#1D for share listing
+  3) WORKGROUP#__ for workgroup listing
+share is the share on the server to query
+path is the SMB path on the server
+return the full path (ie. add cwd if needed)
 *******************************************************/
 char *smbw_parse_path(const char *fname, char *server, char *share, char *path)
 {
 	static pstring s;
-	char *p, *p2;
-	int len = strlen(smbw_prefix)-1;
+	const char *p;
+	int len;
+	fstring workgroup;
 
-	*server = *share = *path = 0;
-
-	if (fname[0] == '/') {
-		pstrcpy(s, fname);
+	/* add cwd if necessary */
+	if (fname[0] != '/') {
+		slprintf(s, sizeof(s), "%s/%s", smbw_cwd, fname);
 	} else {
-		slprintf(s,sizeof(s)-1, "%s/%s", smbw_cwd, fname);
+		pstrcpy(s, fname);
 	}
 	clean_fname(s);
 
-	DEBUG(5,("cleaned %s (fname=%s cwd=%s)\n", 
-		 s, fname, smbw_cwd));
-
+	/* see if it has the right prefix */
+	len = strlen(smbw_prefix)-1;
 	if (strncmp(s,smbw_prefix,len) || 
 	    (s[len] != '/' && s[len] != 0)) return s;
 
-	p = s + len;
+	/* ok, its for us. Now parse out the workgroup, share etc. */
+	p = s+len;
 	if (*p == '/') p++;
-
-	p2 = strchr(p,'/');
-
-	if (p2) {
-		len = (int)(p2-p);
-	} else {
-		len = strlen(p);
-	}
-
-	len = MIN(len,sizeof(fstring)-1);
-
-	strncpy(server, p, len);
-	server[len] = 0;		
-	unix_to_dos(server, server);
-
-	p = p2;
-	if (!p) {
-		if (len == 0) {
-			char *workgroup = smbw_getshared("WORKGROUP");
-			if (!workgroup) workgroup = lp_workgroup();
-			slprintf(server,sizeof(fstring)-1, "%s#1D", workgroup);
-		}
+	if (!next_token(&p, workgroup, "/", sizeof(fstring))) {
+		/* we're in /smb - give a list of workgroups */
+		slprintf(server,sizeof(fstring), "%s#01", smbw_find_workgroup());
 		fstrcpy(share,"IPC$");
 		pstrcpy(path,"");
-		goto ok;
+		return s;
 	}
+	unix_to_dos(workgroup, workgroup);
 
-	p++;
-	p2 = strchr(p,'/');
-
-	if (p2) {
-		len = (int)(p2-p);
-	} else {
-		len = strlen(p);
+	if (!next_token(&p, server, "/", sizeof(fstring))) {
+		/* we are in /smb/WORKGROUP */
+		slprintf(server,sizeof(fstring), "%s#1D", workgroup);
+		fstrcpy(share,"IPC$");
+		pstrcpy(path,"");
+		return s;
 	}
+	unix_to_dos(server, server);
 
-	len = MIN(len,sizeof(fstring)-1);
-	
-	strncpy(share, p, len);
-	share[len] = 0;
+	if (!next_token(&p, share, "/", sizeof(fstring))) {
+		/* we are in /smb/WORKGROUP/SERVER */
+		fstrcpy(share,"IPC$");
+		pstrcpy(path,"");
+		return s;
+	}
 	unix_to_dos(share, share);
-
-	p = p2;
-	if (!p) {
-		pstrcpy(path,"\\");
-		goto ok;
-	}
-
-	unix_to_dos(path,p);
+	unix_to_dos(path, p);
 
 	all_string_sub(path, "/", "\\", 0);
 
- ok:
-	DEBUG(4,("parsed path name=%s cwd=%s [%s] [%s] [%s]\n", 
-		 fname, smbw_cwd,
-		 server, share, path));
-
 	return s;
 }
 
@@ -379,17 +433,51 @@ return a unix errno from a SMB error pair
 *******************************************************/
 int smbw_errno(struct cli_state *c)
 {
-	uint8 eclass;
-	uint32 ecode;
-	int ret;
+	return cli_errno(c);
+}
 
-	ret = cli_error(c, &eclass, &ecode, NULL);
+/* Return a username and password given a server and share name */
 
-	if (ret) {
-		DEBUG(3,("smbw_error %d %d (0x%x) -> %d\n", 
-			 (int)eclass, (int)ecode, (int)ecode, ret));
+void get_envvar_auth_data(char *server, char *share, char **workgroup,
+			  char **username, char **password)
+{
+	static pstring envuser, envpass;
+	/* Fall back to shared memory/environment variables */
+
+	*envpass = '\0';
+	*username = smbw_getshared("USER");
+	if (!*username) {
+		*username = getenv("USER");
+		if (!*username) *username = "guest";
+		else {
+			char *p;
+
+			unix_to_dos(envuser, *username);
+			if ((p=strchr(envuser,'%'))) {
+				*p = 0;
+				pstrcpy(envpass,p+1);
+				memset(strchr(getenv("USER"),'%')+1,'X',strlen(envpass));
+			}
+			strupper(envuser);
+			*username = envuser;
+		}
 	}
-	return ret;
+
+	*workgroup = smbw_getshared("WORKGROUP");
+	if (!*workgroup) *workgroup = lp_workgroup();
+
+	*password = smbw_getshared("PASSWORD");
+	if (!*password) *password = envpass;
+}
+
+static smbw_get_auth_data_fn get_auth_data_fn = get_envvar_auth_data;
+
+/*****************************************************
+set the get auth data function
+******************************************************/
+void smbw_set_auth_data_fn(smbw_get_auth_data_fn fn)
+{
+	get_auth_data_fn = fn;
 }
 
 /***************************************************** 
@@ -407,36 +495,19 @@ struct smbw_server *smbw_server(char *server, char *share)
 	fstring group;
 	pstring ipenv;
 	struct in_addr ip;
-	extern struct in_addr ipzero;
-	pstring envuser, envpass;
 
-	ip = ipzero;
+	zero_ip(&ip);
 	ZERO_STRUCT(c);
 
-	*envpass = '\0';
-	username = smbw_getshared("USER");
-	if (!username && getenv("USER")) {
-		unix_to_dos(envuser,getenv("USER"));
-		if ((p=strchr(envuser,'%'))) {
-			*p = 0;
-			pstrcpy(envpass,p+1);
-			memset(strchr(getenv("USER"),'%')+1,'X',strlen(envpass));
-		}
-		strupper(envuser);
-		username = envuser;
-	}
-	if (!username) username = "GUEST";
-
-	workgroup = smbw_getshared("WORKGROUP");
-	if (!workgroup) workgroup = lp_workgroup();
-
-	password = smbw_getshared("PASSWORD");
-	if (!password) password = envpass;
+	get_auth_data_fn(server, share, &workgroup, &username, &password);
 
 	/* try to use an existing connection */
 	for (srv=smbw_srvs;srv;srv=srv->next) {
 		if (strcmp(server,srv->server_name)==0 &&
-		    strcmp(share,srv->share_name)==0) return srv;
+		    strcmp(share,srv->share_name)==0 &&
+		    strcmp(workgroup,srv->workgroup)==0 &&
+		    strcmp(username, srv->username) == 0) 
+			return srv;
 	}
 
 	if (server[0] == 0) {
@@ -449,7 +520,8 @@ struct smbw_server *smbw_server(char *server, char *share)
 
 	DEBUG(4,("server_n=[%s] server=[%s]\n", server_n, server));
 
-	if ((p=strchr(server_n,'#')) && strcmp(p+1,"1D")==0) {
+	if ((p=strchr(server_n,'#')) && 
+	    (strcmp(p+1,"1D")==0 || strcmp(p+1,"01")==0)) {
 		struct in_addr sip;
 		pstring s;
 
@@ -475,7 +547,7 @@ struct smbw_server *smbw_server(char *server, char *share)
  again:
 	slprintf(ipenv,sizeof(ipenv)-1,"HOST_%s", server_n);
 
-	ip = ipzero;
+	zero_ip(&ip);
 	if ((p=smbw_getshared(ipenv))) {
 		ip = *(interpret_addr2(p));
 	}
@@ -552,6 +624,18 @@ struct smbw_server *smbw_server(char *server, char *share)
 		goto failed;
 	}
 
+	srv->workgroup = strdup(workgroup);
+	if (!srv->workgroup) {
+		errno = ENOMEM;
+		goto failed;
+	}
+
+	srv->username = strdup(username);
+	if (!srv->username) {
+		errno = ENOMEM;
+		goto failed;
+	}
+
 	/* some programs play with file descriptors fairly intimately. We
 	   try to get out of the way by duping to a high fd number */
 	if (fcntl(SMBW_CLI_FD + srv->cli.fd, F_GETFD) && errno == EBADF) {
@@ -570,9 +654,9 @@ struct smbw_server *smbw_server(char *server, char *share)
 	cli_shutdown(&c);
 	if (!srv) return NULL;
 
-	if (srv->server_name) free(srv->server_name);
-	if (srv->share_name) free(srv->share_name);
-	free(srv);
+	SAFE_FREE(srv->server_name);
+	SAFE_FREE(srv->share_name);
+	SAFE_FREE(srv);
 	return NULL;
 }
 
@@ -686,12 +770,10 @@ int smbw_open(const char *fname, int flags, mode_t mode)
 	}
 	if (file) {
 		if (file->f) {
-			if (file->f->fname) {
-				free(file->f->fname);
-			}
-			free(file->f);
+			SAFE_FREE(file->f->fname);
+			SAFE_FREE(file->f);
 		}
-		free(file);
+		SAFE_FREE(file);
 	}
 	smbw_busy--;
 	return -1;
@@ -701,10 +783,10 @@ int smbw_open(const char *fname, int flags, mode_t mode)
 /***************************************************** 
 a wrapper for pread()
 *******************************************************/
-ssize_t smbw_pread(int fd, void *buf, size_t count, off_t ofs)
+ssize_t smbw_pread(int fd, void *buf, size_t count, SMB_OFF_T ofs)
 {
 	struct smbw_file *file;
-	int ret;
+	ssize_t ret;
 
 	smbw_busy++;
 
@@ -733,7 +815,7 @@ a wrapper for read()
 ssize_t smbw_read(int fd, void *buf, size_t count)
 {
 	struct smbw_file *file;
-	int ret;
+	ssize_t ret;
 
 	DEBUG(4,("smbw_read(%d, %d)\n", fd, (int)count));
 
@@ -771,7 +853,7 @@ a wrapper for write()
 ssize_t smbw_write(int fd, void *buf, size_t count)
 {
 	struct smbw_file *file;
-	int ret;
+	ssize_t ret;
 
 	smbw_busy++;
 
@@ -799,10 +881,10 @@ ssize_t smbw_write(int fd, void *buf, size_t count)
 /***************************************************** 
 a wrapper for pwrite()
 *******************************************************/
-ssize_t smbw_pwrite(int fd, void *buf, size_t count, off_t ofs)
+ssize_t smbw_pwrite(int fd, void *buf, size_t count, SMB_OFF_T ofs)
 {
 	struct smbw_file *file;
-	int ret;
+	ssize_t ret;
 
 	smbw_busy++;
 
@@ -856,11 +938,11 @@ int smbw_close(int fd)
 
 	file->f->ref_count--;
 	if (file->f->ref_count == 0) {
-		free(file->f->fname);
-		free(file->f);
+		SAFE_FREE(file->f->fname);
+		SAFE_FREE(file->f);
 	}
 	ZERO_STRUCTP(file);
-	free(file);
+	SAFE_FREE(file);
 	
 	smbw_busy--;
 
@@ -1090,10 +1172,9 @@ int smbw_utime(const char *fname, void *buf)
 /***************************************************** 
 a wrapper for utime 
 *******************************************************/
-int smbw_utimes(const char *fname, void *buf)
+int smbw_utimes(const char *fname, const struct timeval *buf)
 {
-	struct timeval *tbuf = (struct timeval *)buf;
-	return smbw_settime(fname, tbuf?tbuf->tv_sec:time(NULL));
+	return smbw_settime(fname, buf?buf->tv_sec:time(NULL));
 }
 
 
@@ -1196,7 +1277,7 @@ a wrapper for lseek()
 off_t smbw_lseek(int fd, off_t offset, int whence)
 {
 	struct smbw_file *file;
-	size_t size;
+	SMB_BIG_UINT size;
 
 	smbw_busy++;
 
@@ -1348,18 +1429,31 @@ static void smbw_srv_close(struct smbw_server *srv)
 
 	cli_shutdown(&srv->cli);
 
-	free(srv->server_name);
-	free(srv->share_name);
+	SAFE_FREE(srv->server_name);
+	SAFE_FREE(srv->share_name);
 
 	DLIST_REMOVE(smbw_srvs, srv);
 
 	ZERO_STRUCTP(srv);
 
-	free(srv);
+	SAFE_FREE(srv);
 	
 	smbw_busy--;
 }
 
+/*****************************************************
+alternative fork for some broken syscall(SYS_fork)
+*******************************************************/
+static pid_t fixed_fork(void)
+{
+	pid_t pid, child;
+
+	pid = getpid();
+	child = real_fork();
+	if (child == pid) return 0;
+	return child;
+}
+
 /***************************************************** 
 when we fork we have to close all connections and files
 in the child
@@ -1374,9 +1468,9 @@ int smbw_fork(void)
 	struct smbw_file *file, *next_file;
 	struct smbw_server *srv, *next_srv;
 
-	if (pipe(p)) return real_fork();
+	if (pipe(p)) return fixed_fork();
 
-	child = real_fork();
+	child = fixed_fork();
 
 	if (child) {
 		/* block the parent for a moment until the sockets are
@@ -1387,6 +1481,9 @@ int smbw_fork(void)
 		return child;
 	}
 
+	slprintf(line,sizeof(line)-1,"PWD_%d", (int)getpid());
+	smbw_setshared(line,smbw_cwd);
+
 	close(p[0]);
 
 	/* close all files */
@@ -1401,9 +1498,6 @@ int smbw_fork(void)
 		smbw_srv_close(srv);
 	}
 
-	slprintf(line,sizeof(line)-1,"PWD_%d", (int)getpid());
-	smbw_setshared(line,smbw_cwd);
-
 	/* unblock the parent */
 	write(p[1], &c, 1);
 	close(p[1]);
@@ -1436,11 +1530,13 @@ say no to acls
 }
 #endif
 
-#ifdef HAVE_EXPLICIT_LARGEFILE_SUPPORT
 #ifdef HAVE_STAT64
+#if defined(HAVE_STRUCT_STAT64)
 /* this can't be in wrapped.c because of include conflicts */
  void stat64_convert(struct stat *st, struct stat64 *st64)
 {
+	ZERO_STRUCTP(st64);
+
 	st64->st_size = st->st_size;
 	st64->st_mode = st->st_mode;
 	st64->st_ino = st->st_ino;
@@ -1455,72 +1551,194 @@ say no to acls
 	st64->st_blksize = st->st_blksize;
 	st64->st_blocks = st->st_blocks;
 }
+#endif
+#endif
+
+#ifdef HAVE_READDIR64
+#if defined(HAVE_STRUCT_DIRENT64) || defined(HAVE_GLIBC)
+#ifndef HAVE_STRUCT_DIRENT64
+struct dirent64 {
+	__ino64_t d_ino;
+#ifdef HAVE_DIRENT_D_OFF
+	__off64_t d_off;
+#endif
+#ifdef HAVE_DIRENT_D_RECLEN
+	u_short d_reclen;
+#endif
+#ifdef HAVE_DIRENT_D_NAMLEN
+	u_short d_namlen;
+#endif
+#ifdef HAVE_DIRENT_D_TYPE
+	u_char d_type;
+#endif
+	char d_name[MAXNAMLEN + 1];
+};
 #endif
 
-#ifdef HAVE_READDIR64 
  void dirent64_convert(struct dirent *d, struct dirent64 *d64)
 {
 	d64->d_ino = d->d_ino;
+#ifdef HAVE_DIRENT_D_OFF
 	d64->d_off = d->d_off;
+#endif
+#ifdef HAVE_DIRENT_D_RECLEN
 	d64->d_reclen = d->d_reclen;
+#endif
+#ifdef HAVE_DIRENT_D_NAMLEN
+	d64->d_namlen = d->d_namlen;
+#endif
+#ifdef HAVE_DIRENT_D_TYPE
+	d64->d_type = d->d_type;
+#endif
 	pstrcpy(d64->d_name, d->d_name);
 }
+
+ void dirent64_revert(struct dirent64 *d64, struct dirent *d)
+{
+	d->d_ino = d64->d_ino;
+#ifdef HAVE_DIRENT_D_OFF
+	d->d_off = d64->d_off;
+#endif
+#ifdef HAVE_DIRENT_D_RECLEN
+	d->d_reclen = d64->d_reclen;
+#endif
+#ifdef HAVE_DIRENT_D_NAMLEN
+	d->d_namlen = d64->d_namlen;
+#endif
+#ifdef HAVE_DIRENT_D_TYPE
+	d->d_type = d64->d_type;
+#endif
+	pstrcpy(d->d_name, d64->d_name);
+}
 #endif
 #endif
 
 
 #ifdef HAVE___XSTAT
-/* Definition of `struct stat' used in the linux kernel..  */
-struct kernel_stat {
-	unsigned short int st_dev;
-	unsigned short int __pad1;
-	unsigned long int st_ino;
-	unsigned short int st_mode;
-	unsigned short int st_nlink;
-	unsigned short int st_uid;
-	unsigned short int st_gid;
-	unsigned short int st_rdev;
-	unsigned short int __pad2;
-	unsigned long int st_size;
-	unsigned long int st_blksize;
-	unsigned long int st_blocks;
-	unsigned long int st_atime;
-	unsigned long int __unused1;
-	unsigned long int st_mtime;
-	unsigned long int __unused2;
-	unsigned long int st_ctime;
-	unsigned long int __unused3;
-	unsigned long int __unused4;
-	unsigned long int __unused5;
+#ifdef _STAT_VER_GLIBC2
+struct glibc2_stat {
+	__dev_t st_dev;
+	__ino_t st_ino;
+	__mode_t st_mode;
+	__nlink_t st_nlink;
+	__uid_t st_uid;
+	__gid_t st_gid;
+	__dev_t st_rdev;
+	__off_t st_size;
+	__time_t st_atime;
+	__time_t st_mtime;
+	__time_t st_ctime;
+	unsigned int st_blksize;
+	int st_blocks;
+	unsigned int st_flags;
+	unsigned int st_gen;
 };
+#endif
 
-/*
- * Prototype for gcc in 'fussy' mode.
- */
- void xstat_convert(int vers, struct stat *st, struct kernel_stat *kbuf);
- void xstat_convert(int vers, struct stat *st, struct kernel_stat *kbuf)
+void xstat_convert(int vers, void *kbuf, void *ubuf)
 {
+	struct stat *st;
+	struct smbw_xstat xst;
+
 #ifdef _STAT_VER_LINUX_OLD
 	if (vers == _STAT_VER_LINUX_OLD) {
-		memcpy(st, kbuf, sizeof(*st));
+		memcpy(ubuf, kbuf, sizeof(struct stat));
 		return;
 	}
 #endif
 
+	kernel_xstat_convert(kbuf, &xst);
+
+#ifdef _STAT_VER_GLIBC2
+	if (vers == _STAT_VER_GLIBC2) {
+		struct glibc2_stat *buf;
+
+		buf = (struct glibc2_stat *)ubuf;
+		ZERO_STRUCTP(buf);
+
+		buf->st_dev = xst.xst_dev;
+		buf->st_ino = xst.xst_ino;
+		buf->st_mode = xst.xst_mode;
+		buf->st_nlink = xst.xst_nlink;
+		buf->st_uid = xst.xst_uid;
+		buf->st_gid = xst.xst_gid;
+		buf->st_rdev = xst.xst_rdev;
+		buf->st_size = xst.xst_size;
+		buf->st_blksize = xst.xst_blksize;
+		buf->st_blocks = xst.xst_blocks;
+		buf->st_atime = xst.xst_atime;
+		buf->st_mtime = xst.xst_mtime;
+		buf->st_ctime = xst.xst_ctime;
+		return;
+	}
+#endif
+
+#ifdef _STAT_VER_GLIBC2_1
+	if (vers == _STAT_VER_GLIBC2_1) {
+		struct stat64 *buf;
+
+		buf = (struct stat64 *)ubuf;
+		ZERO_STRUCTP(buf);
+
+		buf->st_dev = xst.xst_dev;
+		buf->st_ino = xst.xst_ino;
+		buf->st_mode = xst.xst_mode;
+		buf->st_nlink = xst.xst_nlink;
+		buf->st_uid = xst.xst_uid;
+		buf->st_gid = xst.xst_gid;
+		buf->st_rdev = xst.xst_rdev;
+		buf->st_size = xst.xst_size;
+		buf->st_blksize = xst.xst_blksize;
+		buf->st_blocks = xst.xst_blocks;
+		buf->st_atime = xst.xst_atime;
+		buf->st_mtime = xst.xst_mtime;
+		buf->st_ctime = xst.xst_ctime;
+		return;
+	}
+#endif
+
+	st = (struct stat *)ubuf;
 	ZERO_STRUCTP(st);
 
-	st->st_dev = kbuf->st_dev;
-	st->st_ino = kbuf->st_ino;
-	st->st_mode = kbuf->st_mode;
-	st->st_nlink = kbuf->st_nlink;
-	st->st_uid = kbuf->st_uid;
-	st->st_gid = kbuf->st_gid;
-	st->st_rdev = kbuf->st_rdev;
-	st->st_size = kbuf->st_size;
-	st->st_blksize = kbuf->st_blksize;
-	st->st_blocks = kbuf->st_blocks;
-	st->st_atime = kbuf->st_atime;
-	st->st_mtime = kbuf->st_mtime;
-	st->st_ctime = kbuf->st_ctime;
+	st->st_dev = xst.xst_dev;
+	st->st_ino = xst.xst_ino;
+	st->st_mode = xst.xst_mode;
+	st->st_nlink = xst.xst_nlink;
+	st->st_uid = xst.xst_uid;
+	st->st_gid = xst.xst_gid;
+	st->st_rdev = xst.xst_rdev;
+	st->st_size = xst.xst_size;
+	st->st_blksize = xst.xst_blksize;
+	st->st_blocks = xst.xst_blocks;
+	st->st_atime = xst.xst_atime;
+	st->st_mtime = xst.xst_mtime;
+	st->st_ctime = xst.xst_ctime;
+}
+
+#if defined(HAVE_STRUCT_STAT64)
+void xstat64_convert(int vers, void *kbuf, void *ubuf)
+{
+	struct stat64 *st64;
+	struct smbw_xstat xst;
+
+	kernel_xstat_convert(kbuf, &xst);
+
+	st64 = (struct stat64 *)ubuf;
+	ZERO_STRUCTP(st64);
+
+	st64->st_dev = xst.xst_dev;
+	st64->st_ino = xst.xst_ino;
+	st64->st_mode = xst.xst_mode;
+	st64->st_nlink = xst.xst_nlink;
+	st64->st_uid = xst.xst_uid;
+	st64->st_gid = xst.xst_gid;
+	st64->st_rdev = xst.xst_rdev;
+	st64->st_size = xst.xst_size;
+	st64->st_blksize = xst.xst_blksize;
+	st64->st_blocks = xst.xst_blocks;
+	st64->st_atime = xst.xst_atime;
+	st64->st_mtime = xst.xst_mtime;
+	st64->st_ctime = xst.xst_ctime;
 }
 #endif
+#endif
